__all__ = ('PLUGIN_LOADER', 'PluginLoader', )

from functools import partial as partial_func
from itertools import chain
from os.path import isfile as is_file

from scarletio import (
    CauseGroup, HybridValueDictionary, RichAttributeErrorBaseType, Task, alchemy_incendiary, export,
    is_coroutine_function, run_coroutine, shield
)

from ...discord.core import KOKORO

from .constants import (
    PLUGINS, PLUGIN_ACTION_FLAG_LOAD, PLUGIN_ACTION_FLAG_SYNTAX_CHECK, PLUGIN_ACTION_FLAG_UNLINK,
    PLUGIN_ACTION_FLAG_UNLOAD, PLUGIN_STATE_LOADED
)
from .exceptions import PluginError
from .helpers import (
    PROTECTED_NAMES, _add_plugin_name_to_plugin_root_names, _get_plugin_name_and_path,
    _is_plugin_name_in_plugin_root_names, _try_get_plugin, _validate_entry_or_exit, validate_plugin_parameters
)
from .plugin import Plugin
from .plugin_extractor import PluginExtractor
from .plugin_tree.plugin_tree_helpers import (
    _build_and_sort_plugin_trees, _sort_plugin_trees, _unwrap_plugin_trees_into_plugins
)
from .plugin_tree.plugin_tree_iterator import PluginTreeIterator


def _pop_unloader_task_callback(plugin, task):
    """
    Removes the unloader task of the plugin loader for the given plugin.
    
    Parameters
    ----------
    plugin : ``Plugin``
        The plugin to remove it's task of.
    task : ``Task``
        The finished task.
    """
    try:
        del PLUGIN_LOADER._unloader_tasks[plugin]
    except KeyError:
        pass


def _pop_loader_task_callback(plugin, task):
    """
    Removes the loader task of the plugin loader for the given plugin.
    
    Parameters
    ----------
    plugin : ``Plugin``
        The plugin to remove it's task of.
    task : ``Task``
        The finished task.
    """
    try:
        del PLUGIN_LOADER._loader_tasks[plugin]
    except KeyError:
        pass


def _run_maybe_blocking(coroutine, blocking):
    """
    Starts the given coroutine inside of a task. Returns a wrapped task, or it's result depending from where it was
    called.
    
    Parameters
    ----------
    coroutine : `CoroutineType`
        The coroutine to run.
    blocking : `bool`
        Whether the operation should be blocking when called from a non-async thread.
    
    Returns
    -------
    result : `None`, ``Task``, ``FutureWrapperAsync``
    """
    if blocking:
        return run_coroutine(coroutine, KOKORO)
    else:
        return KOKORO.create_task_thread_safe(coroutine)


class PluginLoader(RichAttributeErrorBaseType):
    """
    There are some cases when you probably want to change some functional part of your client in runtime. Load,
    unload or reload code. Hata provides an easy to use (that's what she said) solution to solve this issue.
    
    It is called plugin loader is a plugin of hata. It is separated from `commands` plugin, but it does not
    mean they do not go well together. But what more, plugin loader was made to complement it.
    
    Usage
    -----
    The ``PluginLoader`` class is instanced already as `PLUGIN_LOADER` and that can be imported as well from
    `plugin_loader.py`. Instancing the class again will yield the same object.
    
    Because hata can have more clients, we needed a special plugin loader what can handle using more clients at any
    file, so the choice ended up on a  really interesting idea: assign variables to a module before it is (re)loaded.
    
    To show this is not blackmagic, here is an example:
    
    **main.py**
    
    ```py
    from hata.ext.plugin_loader import PLUGIN_LOADER
    from datetime import datetime as DateTime, timezone as TimeZone
    
    cake = 'cake'
    now = DateTime.now(TimeZone.utc)
    
    PLUGIN_LOADER.add_default_variables(cake = cake, now = now)
    PLUGIN_LOADER.register_and_load('plugin')
    ```
    
    **plugin.py**
    
    ```py
    print(cake, now)
    ```
    
    Make sure you start `main.py` with interactive mode. If you never did it, just use the `-i` option like:
    
    ```
    $ python3 -i main.py
    ```
    
    Or on windows:
    
    ```
    $ python -i main.py
    ```
    
    After you ran `main.py` you should see the following (the date should be different tho):
    
    ```
    cake 2020-03-14 09:40:41.587673
    ```
    
    Because now we have the interpreter, you can change the variables.
    
    ```py
    >>> PLUGIN_LOADER.add_default_variables(cake = 'cakes taste good, and now is:')
    >>> PLUGIN_LOADER.reload_all()
    ```
    
    And a different text is printed out:
    
    ```
    cakes taste good, and now is: 2020-03-14 09:40:41.587673
    ```
    
    Now lets edit `plugin.py`.
    
    ```py
    cake = cake.split()
    print(*cake, now, sep = '\\n')
    ```
    
    And reload the plugin:
    
    ```py
    >>> PLUGIN_LOADER.reload_all()
    ```
    
    The printed text really changed again:
    
    ```
    cakes
    taste
    good,
    and
    now
    is:
    2020-03-14 09:40:41.587673
    ```
    
    If you remove default variables and the plugin file still uses them, you get an ``PluginError``:
    
    ```py
    >>> PLUGIN_LOADER.remove_default_variables('cake')
    >>> PLUGIN_LOADER.reload_all()
    ```
    
    ```
    Traceback (most recent call last):
      File "<pyshell#13>", line 1, in <module>
        PLUGIN_LOADER.reload_all()
      File ".../hata/ext/plugin_loader/plugin_loader.py", line 652, in reload_all
        task.sync_wrap().wait()
      File ".../scarletio/core/traps/future_sync_wrapper.py", line 823, in wait
        return self.get_result()
      File ".../scarletio/core/traps/future_sync_wrapper.py", line 723, in result
        raise exception
      File ".../scarletio/core/traps/task.py", line 1602, in _step
        result = coroutine.throw(exception)
      File ".../hata/ext/plugin_loader/plugin_loader.py", line 670, in _reload_all_task
        raise PluginError(error_messages) from None
    hata.ext.plugin_loader.PluginError: PluginError (1):
    Exception occurred meanwhile loading a plugin: `plugin`.
    
    Traceback (most recent call last):
      File ".../plugin.py", line 1, in <module>
        cake = cake.split()
    NameError("name 'cake' is not defined")
    ```
    
    Adding Plugins
    -----------------
    Plugins can be register with the `.register` method.
    
    ```py
    PLUGIN_LOADER.register('cute_commands')
    ```
    
    Or more plugin can be added as well by passing an iterable:
    
    ```py
    PLUGIN_LOADER.register(['cute_commands', 'nice_commands'])
    ```
    
    If a plugin's file is not found, then `.register` will raise  `ModuleNotFoundError`. If the passed parameter
    is not `str` or not `iterable` of `str`, `TypeError` is raised.
    
    Loading
    -------
    Plugins can be loaded by their name:
    
    ```py
    PLUGIN_LOADER.load('cute_commands')
    ```
    
    All plugin can be loaded by using:
    
    ```py
    PLUGIN_LOADER.load_all()
    ```
    
    Or plugin can be added and loaded at the same time as well:
    
    ```py
    PLUGIN_LOADER.register_and_load('cute_commands')
    ```
    
    `.register_and_load` method supports all the keyword parameters as `.register`.
    
    ##### Passing variables to plugins
    
    You can pass variables to plugins with the `.add_default_variables` method:
    
    ```py
    PLUGIN_LOADER.add_default_variables(cake = cake, now = now)
    ```
    
    Adding or removing variables wont change the already loaded plugins' state, those needs to be reloaded to see
    them.
    
    Or pass variables to just specific plugins:
    
    ```py
    PLUGIN_LOADER.register('cute_commands', cake = cake, now = now)
    ```
    
    You can specify if the plugin should use just it's own variables and ignore the default ones too:
    
    ```py
    PLUGIN_LOADER.register('cute_commands', extend_default_variables = False, cake = cake, now = now)
    ```
    
    Every variable added is stored in an optionally weak value dictionary, but you are able remove the added variables
    as well:
    
    ```py
    PLUGIN_LOADER.remove_default_variables('cake', 'now')
    ```
    
    The plugins can be accessed by their name as well, then you can use their properties to modify them.
    
    ```py
    PLUGIN_LOADER.get_plugin('cute_commands').remove_default_variables('cake')
    ```
    
    Unloading And Exit Point
    ------------------------
    You can unload plugin on the same way as loading them.
    
    ```py
    PLUGIN_LOADER.unload('cute_commands')
    ```
    
    Or unload all:
    
    ```py
    PLUGIN_LOADER.unload_all()
    ```
    
    When unloading a plugin, the plugin loader will search a function at the plugin, what we call
    `exit_point` and will run it. By default it looks for a variable named `teardown`. `exit_point` acts on the same
    way as the `entry_point`, so it can be modified for looking for other name to defining and passing a callable
    (can be async again).
    
    Can be set almost on the same way as well:
    
    ```py
    PLUGIN_LOADER.default_exit_point = 'exit'
    ```
    
    Or:
    
    ```py
    PLUGIN_LOADER.register('cute_commands', exit_point = 'exit')
    ```
    
    There are also methods for reloading: `.reload(name)` and `.reload_all()`
    
    Removing Plugins
    -------------------
    
    Unloaded plugins can be removed from the plugin loader by using the `.remove` method:
    
    ```py
    PLUGIN_LOADER.remove('cute_commands')
    ```
    
    Or more plugin with an iterable:
    
    ```py
    PLUGIN_LOADER.remove(['cute_commands', 'nice_commands'])
    ```
    
    Removing loaded plugin will yield `RuntimeError`.
    
    Threading Model
    ---------------
    When you call the different methods of the plugin loader, they ll run on the ``Clients`` thread, what is named
    `KOKORO` internally.
    
    These methods are:
    - ``.register_and_load``
    - ``.load``
    - ``.load_all``
    - ``.unload``
    - ``.unload_all``
    - ``.reload``
    - ``.reload_all``
    
    Meanwhile loading and executing the plugin's code the thread is switched to an executor, so blocking tasks can
    be executed easily. The exception under this rule are the `entry_point`-s and the `exit_point`-s, which always run
    on the same thread as the clients, that is why they can be async as well.
    
    These methods also act differently depending from which thread they were called from. Whenever they are called from
    the client's thread, a ``Task`` is returned what can be `awaited`. If called from other ``EventThread``, then the
    task is async wrapped and that is returned. When calling from any other thread (like the main thread for example),
    the task is sync wrapped and the thread is blocked till the plugin's loading is finished.
    
    Attributes
    ----------
    _default_entry_point : `None`, `str`, `callable`
        Internal slot for the ``.default_entry_point`` property.
    _default_exit_point : `None`, `str`, `callable`
        Internal slot for the ``.default_exit_point`` property.
    _default_variables : `None`, `HybridValueDictionary` of (`str`, `object`) items
        An optionally weak value dictionary to store objects for assigning them to modules before loading them.
        If it would be set as empty, then it is set as `None` instead.
    _done_callbacks : `None`, `list` of `callable`
        Callbacks, which run when loading / unloading the plugins is done.
        
        > These callbacks are only called once, and then cleared out.
    _execute_counter : `int`
        Whether the plugin loader is executing a plugin.
    _loader_tasks : `dict` of (``Plugin``, ``Task``) items
        Active plugin loading tasks.
    _plugins_by_name : `dict` of (`str`, ``Plugin``) items
        A dictionary of the added plugins to the plugin loader in `plugin-name` - ``Plugin`` relation.
    _previously_failed_load_trees : `None`, `set` of ``PluginTree``
        Previously failed plugin trees on load (and syntax check).
    _unloader_tasks : `dict` of (``Plugin``, ``Task``) items
        Active plugin unloading tasks.
    
    Class Attributes
    ---------------
    _instance : `None`, ``PluginLoader``
        The already created instance of the ``PluginLoader`` if there is any.
    """
    __slots__ = (
        '_default_entry_point', '_default_exit_point', '_default_variables', '_done_callbacks', '_execute_counter',
        '_loader_tasks', '_plugins_by_name', '_previously_failed_load_trees', '_unloader_tasks'
    )
    
    _instance = None
    
    def __new__(cls):
        """
        Creates an ``PluginLoader``. If the `PluginLoader` was instanced already, then returns that
        instead.
        
        Returns
        -------
        self : ``PluginLoader``
        """
        self = cls._instance
        if self is not None:
            return self
        
        self = object.__new__(cls)
        
        self._default_entry_point = 'setup'
        self._default_exit_point = 'teardown'
        self._done_callbacks = None
        self._plugins_by_name = {}
        self._default_variables = HybridValueDictionary()
        self._execute_counter = 0
        self._loader_tasks = {}
        self._unloader_tasks = {}
        self._previously_failed_load_trees = None
        
        cls._instance = self
        return self
    
    
    @property
    def default_entry_point(self):
        """
        Get-set-del descriptor for modifying the plugin loader's default entry point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._default_entry_point
    
    @default_entry_point.setter
    def default_entry_point(self, default_entry_point):
        if not _validate_entry_or_exit(default_entry_point):
            raise TypeError(
                f'`{self.__class__.__name__}.default_entry_point` can be `None`, `str`, `callable`, got '
                f'{default_entry_point.__class__.__name__}; {default_entry_point!r}.'
            )
        
        self._default_entry_point = default_entry_point
    
    @default_entry_point.deleter
    def default_entry_point(self):
        self._default_entry_point = None
    
    
    @property
    def default_exit_point(self):
        """
        Get-set-del descriptor for modifying the plugin loader's default exit point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._default_exit_point
    
    
    @default_exit_point.setter
    def default_exit_point(self, default_exit_point):
        if not _validate_entry_or_exit(default_exit_point):
            raise TypeError(
                f'`{self.__class__.__name__}.default_exit_point` can be `None`, `str`, `callable`, got '
                f'{default_exit_point.__class__.__name__}; {default_exit_point!r}.'
            )
        
        self._default_exit_point = default_exit_point
    
    
    @default_exit_point.deleter
    def default_exit_point(self):
        self._default_exit_point = None
    
    
    def add_default_variables(self, **variables):
        """
        Adds default variables to the plugin loader.
        
        Parameters
        ----------
        **variables : Keyword Parameters
            Variables to assigned to a plugin's module before it is loaded.
        
        Raises
        ------
        ValueError
             If a variable name is would be used, what is `module` attribute.
        """
        default_variables = self._default_variables
        for key, value in variables.items():
            if key in PROTECTED_NAMES:
                raise ValueError(
                    f'The passed {key!r} is a protected variable name of module type.'
                )
            
            default_variables[key] = value
    
    
    def remove_default_variables(self, *names):
        """
        Removes the mentioned default variables of the plugin loader.
        
        If a variable with a specified name is not found, no error is raised.
        
        Parameters
        ----------
        *names : `str`
            Default variable names.
        """
        default_variables = self._default_variables
        for name in names:
            try:
                del default_variables[name]
            except KeyError:
                pass
    
    
    def clear_default_variables(self):
        """
        Removes all the default variables of the plugin loader.
        """
        self._default_variables.clear()
    
    
    def register(self, name, *positional_parameters, **keyword_parameters):
        """
        Registers a plugin.
        
        If the plugin already exists, returns that one.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The plugin's name to load.
        *positional_parameters : Positional parameters
            Additional parameters to create the plugin with.
        **keyword_parameters : Keyword parameters
            Additional parameters to create the plugin with.
        
        Other Parameters
        ----------------
        entry_point : `None`, `str`, `callable`, Optional
            Plugin specific entry point, to use over the plugin loader's default.
        exit_point : `None`, `str`, `callable`, Optional
            Plugin specific exit point, to use over the plugin loader's default.
        locked : `bool`, Optional
            Whether the given plugin(s) should not be affected by `.{}_all` methods.
        take_snapshot_difference : `bool`, Optional
            Whether snapshot feature should be used.
        **variables : Keyword parameters
            Variables to assign to a plugin(s)'s module before they are loaded.
        
        Raises
        ------
        ModuleNotFoundError
            If the given name do not refers to any loadable file.
        TypeError
            - If `entry_point` was not given as `None`, `str`, `callable`.
            - If `entry_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `exit_point` was not given as `None`, `str`, `callable`.
            - If `exit_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `extend_default_variables` was not given as `bool`.
            - If `locked` was not given as `bool`.
            - If `name` was not given as `str`, `iterable` of `str`.
        ValueError
            If a variable name is would be used, what is `module` attribute.
        """
        plugin_names_and_paths = PluginExtractor.from_strings(
            'name', name
        ).get_plugin_names_and_paths(
            register_directories_as_roots = True
        )
        
        entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference, default_variables = \
            validate_plugin_parameters(*positional_parameters, **keyword_parameters)
        
        for plugin_name, plugin_path in plugin_names_and_paths:
            self._register(
                plugin_name, plugin_path, entry_point, exit_point, extend_default_variables, locked,
                take_snapshot_difference, default_variables
            )
    
    
    def _register(
        self,
        plugin_name,
        plugin_path,
        entry_point,
        exit_point,
        extend_default_variables,
        locked,
        take_snapshot_difference,
        default_variables,
    ):
        """
        Adds a plugin to the plugin loader.
        
        If the plugin already exists, returns that one.
        
        Parameters
        ----------
        plugin_name : `None`, `str`
            The plugin's name to add.
        plugin_path : `None`, `str`
            Path of the plugin to add.
        entry_point : `None`, `str`, `callable`
            Plugin specific entry point, to use over the plugin loader's default.
        exit_point : `None`, `str`, `callable`
            Plugin specific exit point, to use over the plugin loader's default.
        extend_default_variables : `bool`
            Whether the plugin should use the loader's default variables or just it's own.
        locked : `bool`
            Whether the given plugin(s) should not be affected by `.{}_all` methods.
        take_snapshot_difference : `bool`
            Whether snapshot feature should be used.
        default_variables : `None`, `HybridValueDictionary` of (`str`, `object`) items
            An optionally weak value dictionary to store objects for assigning them to modules before loading them.
            If would be empty, is set as `None` instead.
        
        Returns
        -------
        plugin : ``Plugin``
            The registered plugin.
        """
        plugin = Plugin(
            plugin_name, plugin_path, entry_point, exit_point, extend_default_variables,
            locked, take_snapshot_difference, default_variables
        )
        
        self._plugins_by_name[plugin.name] = plugin
        self._plugins_by_name[plugin.file_name] = plugin
        
        short_name = plugin.short_name
        if (short_name is not None):
            self._plugins_by_name.setdefault(plugin.short_name, plugin)
        
        return plugin
    
    
    def remove(self, name):
        """
        Removes one or more plugins from the plugin loader.
        
        If any of the plugins is not found, no errors will be raised.
        
        Parameters
        ----------
        name : `None`, `str`, ``Plugin``, ``PluginTree`` `iterable` of (`None`, `str`, ``Plugin``, ``PluginTree``)
            The plugin(s)'s name(s) to remove.
        
        Raises
        ------
        TypeError
            If `name` was not given as `str`, `iterable` of `str`.
        RuntimeError
            If a loaded plugin would be removed.
        """
        plugins = PluginExtractor('name', name).get_plugins(raise_if_cannot_resolve = False)
        
        for plugin in plugins:
            if plugin._state == PLUGIN_STATE_LOADED:
                raise RuntimeError(
                    f'Plugin `{plugin.name!r}` can not be removed meanwhile it is loaded.'
                )
        
        for plugin in plugins:
            plugin._unlink()
    
    
    def register_and_load(self, name, *parameters, blocking = True, **keyword_parameters):
        """
        Registers then loads the plugin.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The plugin's name to load.
        
        *parameters : Parameters
            Additional parameters to create the plugin with.
        
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to create the plugin with.
        
        Other Parameters
        ----------------
        entry_point : `None`, `str`, `callable`, Optional
            Plugin specific entry point, to use over the plugin loader's default.
        exit_point : `None`, `str`, `callable`, Optional
            Plugin specific exit point, to use over the plugin loader's default.
        locked : `bool`, Optional
            Whether the given plugin(s) should not be affected by `.{}_all` methods.
        take_snapshot_difference : `bool`, Optional
            Whether snapshot feature should be used.
        **variables : Keyword parameters
            Variables to assign to a plugin(s)'s module before they are loaded.
        
        Returns
        -------
        task : `set` of ``Plugin``, ``Task`` -> ``Plugin``, ``FutureWrapperAsync`` -> ``Plugin``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            loading is done. However if called from a sync thread, will block till the loading is done.
            
            When finished returns the loaded plugin.
        
        Raises
        ------
        ImportError
            If the given name do not refers to any loadable file.
        TypeError
            - If `entry_point` was not given as `None`, `str`, `callable`.
            - If `entry_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `exit_point` was not given as `None`, `str`, `callable`.
            - If `exit_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `extend_default_variables` was not given as `bool`.
            - If `locked` was not given as `bool`.
            - If `name` was not given as `str`, `iterable` of `str`.
        ValueError
            If a variable name is would be used, what is `module` attribute.
        PluginError
            The plugin failed to load correctly.
        """
        return _run_maybe_blocking(self._load_plugin_task(name, parameters, keyword_parameters), blocking)
    
    
    async def _load_plugin_task(self, name, parameters, keyword_parameters):
        """
        Adds, then loads the plugin.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str`
            The plugin's name.
        parameters : `tuple` of `object`
            Additional parameters to create the plugin with.
        keyword_parameters : `dict<str, object>`
            Additional parameters to create the plugin with.
        
        Returns
        -------
        plugin : ``Plugin``
            The loaded plugin.
        
        Raises
        ------
        ImportError
            If the given name do not refers to any loadable file.
        TypeError
            - If `entry_point` was not given as `None`, `str`, `callable`.
            - If `entry_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `exit_point` was not given as `None`, `str`, `callable`.
            - If `exit_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `extend_default_variables` was not given as `bool`.
            - If `locked` was not given as `bool`.
            - If `name` was not given as `str`, `iterable` of `str`.
        ValueError
            If a variable name is would be used, what is `module` attribute.
        PluginError
            The plugin failed to load correctly.
        """
        plugin_name_and_path_pair = _get_plugin_name_and_path(name, False)
        if plugin_name_and_path_pair is None:
            raise ModuleNotFoundError(
                f'No plugin(s) found with the given name: {name!r}.'
            ) from None
        
        plugin_name, plugin_path = plugin_name_and_path_pair
        
        plugin = _try_get_plugin(plugin_name, plugin_path)
        
        entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference, default_variables = \
            validate_plugin_parameters(*parameters, **keyword_parameters)
    
        if (plugin is None):
            plugin = self._register(
                plugin_name, plugin_path, entry_point, exit_point, extend_default_variables, locked,
                take_snapshot_difference, default_variables
            )
            
        else:
            if (default_variables is not None):
                plugin.add_default_variables(**default_variables)
            
            if plugin.is_loaded():
                return plugin
        
        plugin_error = await self._plugin_loader(plugin)
        if (plugin_error is not None):
            raise plugin_error
        
        return plugin
    
    
    def load(self, name, *, blocking = True, deep = True):
        """
        Loads the plugin with the given name. If the plugin is already loaded, will do nothing.
        
        Parameters
        ----------
        name : `None`, `str`, ``Plugin``, ``PluginTree`` `iterable` of (`None`, `str`, ``Plugin``, ``PluginTree``)
            The plugin's name.
        
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        deep : `bool` = `True`, Optional (Keyword only)
            Whether the plugin with all of it's parent and with their child should be reloaded.
        
        Returns
        -------
        task : `list` of ``Plugin``, ``Task`` -> `list` of ``Plugin``,
                ``FutureWrapperAsync`` -> `list` of ``Plugin``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            loading is done. However if called from a sync thread, will block till the loading is done.
            
            When finished returns the loaded plugin.
        
        Raises
        ------
        PluginError
            - No plugin is added with the given name.
            - Loading the plugin failed.
        """
        return _run_maybe_blocking(self._load_task(name, deep), blocking)
    
    
    async def _load_task(self, name, deep):
        """
        Loads the given plugins.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str'
            The plugin(s) name to load.
        deep : `bool`
            Whether the plugin with all of it's parent and with their child should be reloaded.

        Returns
        -------
        plugins : `list` of ``Plugin``
            The loaded plugins.
        
        Raises
        ------
        PluginError
            - No plugin is added with the given name.
            - Loading the plugin failed.
        """
        plugins = PluginExtractor('name', name).get_plugins()
        plugins, plugin_trees = self._pull_previously_failed_plugin_trees(plugins)
        plugin_trees = _build_and_sort_plugin_trees(plugins, plugin_trees, deep)
        await self._execute_actions_on_plugin_trees(
            plugin_trees,
            PLUGIN_ACTION_FLAG_LOAD | PLUGIN_ACTION_FLAG_SYNTAX_CHECK | PLUGIN_ACTION_FLAG_UNLINK,
        )
        return _unwrap_plugin_trees_into_plugins(plugin_trees)
    
    
    def unload(self, name, *, blocking = True, deep = True):
        """
        Unloads the plugin with the given name.
        
        Parameters
        ----------
        name : `None`, `str`, ``Plugin``, ``PluginTree`` `iterable` of (`None`, `str`, ``Plugin``, ``PluginTree``)
            The plugin's name.
        
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        deep : `bool` = `True`, Optional (Keyword only)
            Whether the plugin with all of it's parent and with their child should be reloaded.
        
        Returns
        -------
        task : `list` of ``Plugin``, ``Task`` -> `list` of ``Plugin``,
                ``FutureWrapperAsync`` -> `list` of ``Plugin``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            unloading is done. However if called from a sync thread, will block till the unloading is done.
            
            When finished returns the unloaded plugin.
        
        Raises
        ------
        PluginError
            - No plugin is added with the given name.
            - Unloading the plugin failed.
        """
        return _run_maybe_blocking(self._unload_task(name, deep), blocking)
    
    
    async def _unload_task(self, name, deep):
        """
        Unloads the plugin with the given name.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The plugin's name.
        deep : `bool`
            Whether the plugin with all of it's parent and with their child should be reloaded.
        
        Returns
        -------
        plugins : `list` of ``Plugin``
            The unloaded plugins.
        
        Raises
        ------
        PluginError
            - No plugin is added with the given name.
            - Unloading the plugin failed.
        """
        plugins = PluginExtractor('name', name).get_plugins()
        plugins, plugin_trees = self._pull_previously_failed_plugin_trees(plugins)
        plugin_trees = _build_and_sort_plugin_trees(plugins, plugin_trees, deep)
        await self._execute_actions_on_plugin_trees(
            plugin_trees,
            PLUGIN_ACTION_FLAG_UNLOAD | PLUGIN_ACTION_FLAG_UNLINK,
        )
        return _unwrap_plugin_trees_into_plugins(plugin_trees)
    
    
    def reload(self, name, *, blocking = True, deep = True):
        """
        Reloads the plugin with the given name.
        
        Parameters
        ----------
        name : `None`, `str`, ``Plugin``, ``PluginTree`` `iterable` of (`None`, `str`, ``Plugin``, ``PluginTree``)
            The plugin's name.
        
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.

        deep : `bool` = `True`, Optional (Keyword only)
            Whether the plugin with all of it's parent and with their child should be reloaded.
        
        Returns
        -------
        task : `list` of ``Plugin``, ``Task`` -> `list` of ``Plugin``,
                ``FutureWrapperAsync`` -> `list` of ``Plugin``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            reloading is done. However if called from a sync thread, will block till the reloading is done.
            
            When finished returns the reloaded plugins.
        
        Raises
        ------
        PluginError
            - No plugin is added with the given name.
            - Reloading the plugin failed.
        """
        return _run_maybe_blocking(self._reload_task(name, deep), blocking)
    
    
    async def _reload_task(self, name, deep):
        """
        Reloads the plugin with the given name.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The plugin's name.
        deep : `bool`
            Whether the plugin with all of it's parent and with their child should be reloaded.
        
        Returns
        -------
        plugins : `list` of ``Plugin``
            The reloaded plugins.
        
        Raises
        ------
        PluginError
            - No plugin is added with the given name.
            - Reloading the plugin failed.
        """
        plugins = PluginExtractor('name', name).get_plugins()
        plugins, plugin_trees = self._pull_previously_failed_plugin_trees(plugins)
        plugin_trees = _build_and_sort_plugin_trees(plugins, plugin_trees, deep)
        await self._execute_actions_on_plugin_trees(
            plugin_trees,
            (
                PLUGIN_ACTION_FLAG_LOAD | PLUGIN_ACTION_FLAG_UNLOAD | PLUGIN_ACTION_FLAG_SYNTAX_CHECK |
                PLUGIN_ACTION_FLAG_UNLINK
            ),
        )
        return _unwrap_plugin_trees_into_plugins(plugin_trees)
    
    
    def load_all(self, *, blocking = True):
        """
        Loads all the plugin of the plugin loader. If anything goes wrong, raises an ``PluginError`` only
        at the end, with the exception(s).
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureWrapperAsync``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            loading is done. However if called from a sync thread, will block till the loading is done.
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        Raises
        ------
        PluginError
            If any plugin failed to load correctly.
        """
        return _run_maybe_blocking(self._load_all_task(), blocking)
    
    
    async def _load_all_task(self):
        """
        Loads all the plugins of the plugin loader.
        
        Loads each plugin one after the other. The raised exceptions' messages are collected into one exception,
        what will be raised only at the end. If any of the plugins raises, will still try to unload the leftover
        ones.
        
        This method is a coroutine.
        
        Raises
        ------
        PluginError
            If any plugin failed to load correctly.
        """
        plugins = [plugin for plugin in PLUGINS.values() if not plugin._locked]
        plugins, plugin_trees = self._pull_previously_failed_plugin_trees(plugins)
        plugin_trees = _build_and_sort_plugin_trees(plugins, plugin_trees, False)
        
        await self._execute_actions_on_plugin_trees(
            plugin_trees,
            PLUGIN_ACTION_FLAG_LOAD | PLUGIN_ACTION_FLAG_SYNTAX_CHECK | PLUGIN_ACTION_FLAG_UNLINK,
        )
    
    
    def unload_all(self, *, blocking = True):
        """
        Unloads all the plugin of the plugin loader. If anything goes wrong, raises an ``PluginError`` only
        at the end, with the exception(s).
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureWrapperAsync``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            unloading is done. However if called from a sync thread, will block till the unloading is done.
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        Raises
        ------
        PluginError
            If any plugin failed to unload correctly.
        """
        return _run_maybe_blocking(self._unload_all_task(), blocking)
    
    
    async def _unload_all_task(self):
        """
        Unloads all the plugins of the plugin loader.
        
        Unloads each plugin one after the other. The raised exceptions' messages are collected into one exception,
        what will be raised only at the end. If any of the plugins raises, will still try to unload the leftover
        ones.
        
        This method is a coroutine.
        
        Raises
        ------
        PluginError
            If any plugin failed to unload correctly.
        """
        plugins = [plugin for plugin in PLUGINS.values() if not plugin._locked]
        plugins, plugin_trees = self._pull_previously_failed_plugin_trees(plugins)
        plugin_trees = _build_and_sort_plugin_trees(plugins, plugin_trees, False)
        
        await self._execute_actions_on_plugin_trees(
            plugin_trees,
            PLUGIN_ACTION_FLAG_UNLOAD | PLUGIN_ACTION_FLAG_UNLINK,
        )
    
    
    def reload_all(self, *, blocking = True):
        """
        Reloads all the plugin of the plugin loader. If anything goes wrong, raises an ``PluginError`` only
        at the end, with the exception(s).
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureWrapperAsync``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            reloading is done. However if called from a sync thread, will block till the reloading is done.
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        Raises
        ------
        PluginError
            If any plugin failed to reload correctly.
        """
        return _run_maybe_blocking(self._reload_all_task(), blocking)
    
    
    async def _reload_all_task(self):
        """
        Reloads all the plugins of the plugin loader.
        
        If a plugin is not unloaded, will load it, if the plugin is loaded, will unload, then load it.
        Reloads each plugin one after the other. The raised exceptions' messages are collected into one exception,
        what will be raised at the end. If any of the plugins raises, will still try to reload the leftover ones.
        
        This method is a coroutine.
        
        Raises
        ------
        PluginError
            If any plugin failed to reload correctly.
        """
        plugins = [plugin for plugin in PLUGINS.values() if not plugin._locked]
        plugins, plugin_trees = self._pull_previously_failed_plugin_trees(plugins)
        plugin_trees = _build_and_sort_plugin_trees(plugins, plugin_trees, False)
        
        await self._execute_actions_on_plugin_trees(
            plugin_trees,
            (
                PLUGIN_ACTION_FLAG_LOAD | PLUGIN_ACTION_FLAG_UNLOAD | PLUGIN_ACTION_FLAG_SYNTAX_CHECK |
                PLUGIN_ACTION_FLAG_UNLINK
            ),
        )
    
    
    async def _execute_actions_on_plugin_trees(self, plugin_trees, actions):
        """
        Executes the actions on the given plugins.
        
        This method is a coroutine.
        
        Parameters
        ----------
        plugin_trees : `list` of ``PluginTree``
            The plugin trees to execute the actions on.
        actions : `int`
            Bitwise flags of the actions to executed.
        
        Raises
        ------
        PluginError
            If any plugin failed to load / reload correctly.
        """
        exceptions = None
        plugin_tree_iterators = None
        
        if (actions & PLUGIN_ACTION_FLAG_SYNTAX_CHECK) and plugin_trees:
            exceptions, plugin_tree_iterators = await self._execute_syntax_check_on_plugin_trees(
                plugin_trees, exceptions, plugin_tree_iterators
            )
            
            if (exceptions is not None) and (plugin_tree_iterators is not None):
                plugin_tree_iterator = plugin_tree_iterators[-1]
                plugin_trees = _sort_plugin_trees({*plugin_tree_iterator.iter_done_success()})
        
        if (actions & PLUGIN_ACTION_FLAG_UNLOAD) and plugin_trees:
            exceptions, plugin_tree_iterators = await self._execute_unload_on_plugin_trees(
                plugin_trees, exceptions, plugin_tree_iterators
            )
        
        if (actions & PLUGIN_ACTION_FLAG_UNLINK) and plugin_trees:
            plugin_trees, plugin_tree_iterators = self._execute_unlink_on_plugin_trees(
                plugin_trees, plugin_tree_iterators
            )
            
        if (actions & PLUGIN_ACTION_FLAG_LOAD) and plugin_trees:
            exceptions, plugin_tree_iterators = await self._execute_load_on_plugin_trees(
                plugin_trees, exceptions, plugin_tree_iterators
            )
        
        if (exceptions is not None):
            self._push_previously_failed_plugin_trees_from_iterators(plugin_tree_iterators)
            
            try:
                raise PluginError(
                    action = actions, cause = CauseGroup(*exceptions), plugin_tree_iterators = plugin_tree_iterators
                )
            finally:
                exceptions = None
    
    
    def _execute_unlink_on_plugin_trees(self, plugin_trees, plugin_tree_iterators):
        """
        Unlinks the deleted plugins from the plugin trees.
        
        Parameters
        ----------
        plugin_trees : `list` of ``PluginTree``
            The plugin trees to unload.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        
        Returns
        -------
        plugin_trees : `list` of ``PluginTree``
            The new plugin trees.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        """
        if plugin_tree_iterators is None:
            plugin_tree_iterators = []
        
        plugin_tree_iterator = PluginTreeIterator(plugin_trees, PLUGIN_ACTION_FLAG_UNLINK)
        plugin_tree_iterators.append(plugin_tree_iterator)
        
        to_unlink = None
        plugin_root_names = None
        
        for plugin in plugin_tree_iterator:
            if plugin._state == PLUGIN_STATE_LOADED or is_file(plugin.path):
                continue
            
            plugin_root_names = _add_plugin_name_to_plugin_root_names(plugin_root_names, plugin.name)
            
            if to_unlink is None:
                to_unlink = set()
            
            to_unlink.add(plugin)
            continue
        
        if (to_unlink is not None):
            # Collect additional plugins that are under in the tree than ours.
            for plugin in PLUGINS.values():
                if plugin in to_unlink:
                    continue
                    
                if not _is_plugin_name_in_plugin_root_names(plugin_root_names, plugin.name):
                    continue
                
                if plugin._state == PLUGIN_STATE_LOADED or is_file(plugin.path):
                    continue
                
                to_unlink.add(plugin)
                continue
            
            for plugin in to_unlink:
                plugin._unlink()
            
            plugin_trees = [
                plugin_tree for plugin_tree in
                (plugin_tree.copy_without(to_unlink) for plugin_tree in plugin_trees)
                if plugin_tree is not None
            ]
        
        return plugin_trees, plugin_tree_iterators
    
    
    async def _execute_load_on_plugin_trees(self, plugin_trees, exceptions, plugin_tree_iterators):
        """
        Executes load on the given plugins
        
        This method is a coroutine.
        
        Parameters
        ----------
        plugin_trees : `list` of ``PluginTree``
            The plugin trees to unload.
        exceptions : `None`, `list` of `BaseException`
            Occurred exceptions.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        
        Returns
        -------
        exceptions : `None`, `list` of `BaseException`
            Occurred exceptions.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        """
        if plugin_tree_iterators is None:
            plugin_tree_iterators = []
        
        plugin_tree_iterator = PluginTreeIterator(plugin_trees, PLUGIN_ACTION_FLAG_LOAD)
        plugin_tree_iterators.append(plugin_tree_iterator)
        
        for plugin in plugin_tree_iterator:
            exception = await self._plugin_loader(plugin)
            if (exception is not None):
                plugin_tree_iterator.fail_current_hard()
                plugin_tree_iterator.cancel_dependents_in_to_do(exception.get_plugins())
                if exceptions is None:
                    exceptions = []
                
                exceptions.append(exception)
        
        return exceptions, plugin_tree_iterators
    
    
    async def _execute_unload_on_plugin_trees(self, plugin_trees, exceptions, plugin_tree_iterators):
        """
        Executes unload on the given plugins
        
        This method is a coroutine.
        
        Parameters
        ----------
        plugin_trees : `list` of ``PluginTree``
            The plugin trees to unload.
        exceptions : `None`, `list` of `BaseException`
            Occurred exceptions.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        
        Returns
        -------
        exceptions : `None`, `list` of `BaseException`
            Occurred exceptions.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        """
        if plugin_tree_iterators is None:
            plugin_tree_iterators = []
        
        plugin_tree_iterator = PluginTreeIterator(plugin_trees, PLUGIN_ACTION_FLAG_UNLOAD)
        plugin_tree_iterators.append(plugin_tree_iterator)
        
        for plugin in plugin_tree_iterator:
            exception = await self._plugin_unloader(plugin)
            if (exception is not None):
                plugin_tree_iterator.fail_current_soft()
                
                if exceptions is None:
                    exceptions = []
                
                exceptions.append(exception)
        
        return exceptions, plugin_tree_iterators
    
    
    async def _plugin_loader(self, plugin):
        """
        Loads the given plugin. This method synchronises plugin loading.
        
        This method is a coroutine.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to unload.
        
        Returns
        -------
        exception : `None`, ``PluginError``
        """
        try:
            unloader_task = self._unloader_tasks[plugin]
        except KeyError:
            pass
        else:
            try:
                exception = await shield(unloader_task, KOKORO)
            except:
                raise
            else:
                if (exception is not None):
                    return exception
            
            finally:
                unloader_task = None
            
        try:
            loader_task = self._loader_tasks[plugin]
        except KeyError:
            loader_task = Task(KOKORO, self._plugin_loader_task(plugin))
            loader_task.add_done_callback(partial_func(_pop_loader_task_callback, plugin))
            
            self._loader_tasks[plugin] = loader_task
        
        try:
            return await shield(loader_task, KOKORO)
        finally:
            loader_task = None
    
    
    async def _plugin_loader_task(self, plugin):
        """
        Loads the plugin. If the plugin is loaded, will do nothing.
        
        Loading an exception can be separated to 4 parts:
        
        - Assign the default variables.
        - Load the module.
        - Find the entry point (if needed).
        - Ensure the entry point (if found).
        
        If any of these fails, an ``PluginError`` will be raised. If step 1 raises, then a traceback will be
        included as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to load.
        
        Returns
        -------
        exception : `None`, ``PluginError``
        """
        self._execute_counter += 1
        try:
            try:
                # loading blocks, but unloading does not
                module = await KOKORO.run_in_executor(plugin._load)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                return PluginError(
                    f'Exception occurred meanwhile loading a plugin: `{plugin.name}`',
                    action = PLUGIN_ACTION_FLAG_LOAD,
                    cause = err,
                    plugin = plugin,
                )
            
            if module is None:
                return # already loaded
            
            entry_point = plugin._entry_point
            if entry_point is None:
                entry_point = self._default_entry_point
                if entry_point is None:
                    return
            
            if isinstance(entry_point, str):
                entry_point = getattr(module, entry_point, None)
                if entry_point is None:
                    return
            
            try:
                if is_coroutine_function(entry_point):
                    await entry_point(module)
                else:
                    entry_point(module)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                return PluginError(
                    (
                        f'Exception occurred meanwhile entering a plugin: `{plugin.name}`; '
                        f'At entry_point: {entry_point!r}.',
                    ),
                    cause = err,
                )
        
        
        finally:
            execute_counter = self._execute_counter - 1
            self._execute_counter = execute_counter
            if not execute_counter:
                self.call_done_callbacks()
    
    
    async def _plugin_unloader(self, plugin):
        """
        Unloads the given plugin. This method synchronises plugin unloading.
        
        This method is a coroutine.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to unload.
        
        Returns
        -------
        exception : `None`, ``PluginError``
        """
        try:
            loader_task = self._loader_tasks[plugin]
        except KeyError:
            pass
        else:
            try:
                exception = await shield(loader_task, KOKORO)
            except:
                raise
            else:
                if (exception is not None):
                    return exception
            
            finally:
                loader_task = None
        
        try:
            unloader_task = self._unloader_tasks[plugin]
        except KeyError:
            unloader_task = Task(KOKORO, self._plugin_unloader_task(plugin))
            unloader_task.add_done_callback(partial_func(_pop_unloader_task_callback, plugin, ))
            
            self._unloader_tasks[plugin] = unloader_task
        
        try:
            return await shield(unloader_task, KOKORO)
        finally:
            unloader_task = None

    
    async def _plugin_unloader_task(self, plugin):
        """
        Unloads the plugin. If the plugin is not loaded, will do nothing.
        
        Loading an exception can be separated to 3 parts:
        
        - Find the exit point (if needed).
        - Ensure the exit point (if found).
        - Remove the default variables.
        
        If any of these fails, an ``PluginError`` will be raised. If step 2 raises, then a traceback will be
        included as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to unload.
        
        Returns
        -------
        exception : `None`, ``PluginError``
        """
        self._execute_counter += 1
        try:
            # loading blocks, but unloading does not
            module = plugin._unload()
            
            if module is None:
                return # not loaded
            
            try:
                exit_point = plugin._exit_point
                if exit_point is None:
                    exit_point = self._default_exit_point
                    if exit_point is None:
                        return
                
                if isinstance(exit_point, str):
                    exit_point = getattr(module, exit_point, None)
                    if exit_point is None:
                        return
                
                try:
                    
                    if is_coroutine_function(exit_point):
                        await exit_point(module)
                    else:
                        exit_point(module)
                except GeneratorExit:
                    raise
                
                except BaseException as err:
                    return PluginError(
                        (
                            f'Exception occurred meanwhile unloading a plugin: `{plugin.name}`; '
                            f'At exit point: {exit_point!r},'
                        ),
                        action = PLUGIN_ACTION_FLAG_UNLOAD,
                        cause = err,
                        plugin = plugin,
                    )
            
            finally:
                plugin._unassign_variables()
                
                keys = []
                module_globals = module.__dict__
                for key in module_globals:
                    if (not key.startswith('__')) and (not key.endswith('__')):
                        keys.append(key)
                
                for key in keys:
                    del module_globals[key]
        finally:
            execute_counter = self._execute_counter - 1
            self._execute_counter = execute_counter
            if not execute_counter:
                self.call_done_callbacks()
    
    
    def __repr__(self):
        """Returns the plugin loader's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' plugin count: ',
            repr(len(PLUGINS)),
        ]
        
        entry_point = self._default_entry_point
        if (entry_point is not None):
            repr_parts.append(', default_entry_point = ')
            repr_parts.append(repr(entry_point))
        
        exit_point = self._default_exit_point
        if (exit_point is not None):
            repr_parts.append(', default_exit_point = ')
            repr_parts.append(repr(exit_point))
        
        default_variables = self._default_variables
        if default_variables:
            repr_parts.append(', default_variables = ')
            repr_parts.append(repr(default_variables))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def is_processing_plugin(self):
        """
        Returns whether the plugin loader is processing a plugin.
        
        Returns
        -------
        is_processing_plugin : `bool`
        """
        if self._execute_counter:
            is_processing_plugin = True
        else:
            is_processing_plugin = False
        
        return is_processing_plugin
    
    
    def get_plugin(self, name):
        """
        Returns the plugin with the given name. The plugin must be already registered.
        
        Parameters
        ----------
        name : `str`
            An plugin's name or path.
        
        Returns
        -------
        plugin : ``Plugin``, `None`.
            The matched plugin if any.
        """
        plugin_name_and_path_pair = _get_plugin_name_and_path(name, True)
        if (plugin_name_and_path_pair is not None):
            return _try_get_plugin(*plugin_name_and_path_pair)
    
    
    async def _execute_syntax_check_on_plugin_trees(self, plugin_trees, exceptions, plugin_tree_iterators):
        """
        Check syntax of the given plugin trees.
        
        This function is a coroutine.
        
        Parameters
        ----------
        plugin_trees : `list` of ``PluginTree``
            A list of plugin trees to check their syntax.
        exceptions : `None`, `list` of `BaseException`
            Occurred exceptions.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        
        Returns
        -------
        exceptions : `None`, `list` of `BaseException`
            Occurred exceptions.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        """
        return await KOKORO.run_in_executor(alchemy_incendiary(
            self._execute_syntax_check_on_plugin_trees_blocking, (plugin_trees, exceptions, plugin_tree_iterators)
        ))
    
    
    def _execute_syntax_check_on_plugin_trees_blocking(self, plugin_trees, exceptions, plugin_tree_iterators):
        """
        Checks whether the plugins can be reloaded.
        
        This method is blocking and ran inside of an executor by ``._execute_syntax_check_on_plugin_trees``.
        
        Parameters
        ----------
        plugin-trees : `list` of ``PluginTree``
            A list of plugin trees to check their syntax.
        exceptions : `None`, `list` of `BaseException`
            Occurred exceptions.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        
        Returns
        -------
        exceptions : `None`, `list` of `BaseException`
            Occurred exceptions.
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`
            Plugin tree iterators.
        """
        if plugin_tree_iterators is None:
            plugin_tree_iterators = []
        
        plugin_tree_iterator = PluginTreeIterator(plugin_trees, PLUGIN_ACTION_FLAG_SYNTAX_CHECK)
        plugin_tree_iterators.append(plugin_tree_iterator)
            
        for plugin in plugin_tree_iterator:
            exception = plugin._check_for_syntax()
            if (exception is not None):
                plugin_tree_iterator.fail_current_hard()
                plugin_tree_iterator.cancel_dependents_in_to_do(exception.get_plugins())
                if exceptions is None:
                    exceptions = []
                
                exceptions.append(exception)
        
        return exceptions, plugin_tree_iterators
    
    
    def add_done_callback(self, callback):
        """
        Adds a done callback to be called when the loading / unloading plugins is finished.
        
        These callbacks are only called once, and then cleared out.
        
        Parameters
        ----------
        callback : `callable`
        
        Returns
        -------
        added : `bool`
            Whether the callback was added.
        """
        done_callbacks = self._done_callbacks
        if (done_callbacks is None):
            done_callbacks = []
            self._done_callbacks = done_callbacks
        
        done_callbacks.append(callback)
        return True
    
    
    def add_done_callback_unique(self, callback):
        """
        Adds a done callback to be called when the loading / unloading plugins is finished.
        
        > Ff the callback is already added, does nothing.
        
        These callbacks are only called once, and then cleared out.
        
        Parameters
        ----------
        callback : `callable`
        
        Returns
        -------
        added : `bool`
            Whether the callback was added.
        """
        done_callbacks = self._done_callbacks
        if (done_callbacks is None):
            done_callbacks = []
            self._done_callbacks = done_callbacks
        
        else:
            if (callback in done_callbacks):
                return False
        
        done_callbacks.append(callback)
        return True
    
    
    def call_done_callbacks(self):
        """
        Calls done callbacks of the plugin loader.
        """
        done_callbacks = self._done_callbacks
        if (done_callbacks is not None):
            self._done_callbacks = done_callbacks
            
            for done_callback in done_callbacks:
                KOKORO.call_soon(done_callback)
    
    
    def _pull_previously_failed_plugin_trees(self, plugins):
        """
        Pulls previously failed plugin trees.
        
        Parameters
        ----------
        plugins : `list` of ``Plugins``
            Plugins to get failed trees for.
        
        Returns
        -------
        plugins : `list` of ``Plugin``
            Plugins excluding the matched ones.
        plugin_trees : `list` of ``PluginTree``
            The matched plugin trees.
        """
        plugin_trees = []
        
        previously_failed_load_trees = self._previously_failed_load_trees
        if previously_failed_load_trees is None:
            return plugins, plugin_trees
        
        found_plugins = set()
        
        for plugin_tree in previously_failed_load_trees:
            intersection = plugin_tree.get_intersection_with_plugins(plugins)
            if not intersection:
                continue
            
            found_plugins.update(intersection)
            plugin_trees.append(plugin_tree)
            continue
        
        if plugin_trees:
            for plugin_tree in plugin_trees:
                previously_failed_load_trees.discard(plugin_tree)
            
            if not previously_failed_load_trees:
                self._previously_failed_load_trees = None
            
            plugins = [plugin for plugin in plugins if plugin not in found_plugins]
        
        return plugins, plugin_trees
    
    
    def _push_previously_failed_plugin_trees_from_iterators(self, plugin_tree_iterators):
        """
        Pushes previous failed plugin trees from plugin tree iterators.
        
        Parameters
        ----------
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator``
            Plugin tree iterators to pull the failed plugin trees from.
        """
        if plugin_tree_iterators is None:
            return
        
        previously_failed_load_trees = self._previously_failed_load_trees
        for plugin_tree_iterator in plugin_tree_iterators:
            if not (plugin_tree_iterator.action & (PLUGIN_ACTION_FLAG_SYNTAX_CHECK | PLUGIN_ACTION_FLAG_LOAD)):
                continue
            
            for plugin_tree in chain(plugin_tree_iterator.iter_done_fail(), plugin_tree_iterator.iter_done_cancelled()):
                if previously_failed_load_trees is None:
                    previously_failed_load_trees = set()
                    self._previously_failed_load_trees = previously_failed_load_trees
                
                previously_failed_load_trees.add(plugin_tree)


PLUGIN_LOADER = PluginLoader()
export(PLUGIN_LOADER, 'PLUGIN_LOADER')



__all__ = ()


from ...field_validators import entity_id_validator_factory, preinstanced_validator_factory
from ...user import ClientUserBase

from ..audit_log.fields import validate_guild_id
from ..audit_log_entry import AuditLogEntryType


# user_id

validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)


# entry_type

validate_entry_type = preinstanced_validator_factory('entry_type', AuditLogEntryType)



__all__ = ('ChannelMetadataGuildThreadPrivate',)

from scarletio import copy_docs

from .fields import parse_invitable, put_invitable, validate_invitable

from .guild_thread_base import ChannelMetadataGuildThreadBase


class ChannelMetadataGuildThreadPrivate(ChannelMetadataGuildThreadBase):
    """
    Guild channel metadata for private threads.
    
    Attributes
    ----------
    _created_at : `None | DateTime`
        When the channel was created.
    _cache_permission : ``None | dict<int, Permission>``
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    archived : `bool`
        Whether the thread s archived.
    archived_at : `None | DateTime`
        When the thread's archive status was last changed.
    auto_archive_after : `int`
        Duration in seconds to automatically archive the thread after recent activity. Can be one of: `3600`, `86400`,
        `259200`, `604800`.
    invitable : `bool`
        Whether non-moderators can invite other non-moderators to the threads.
    name : `str`
        The channel's name.
    open : `bool`
        Whether the thread channel is open.
    owner_id : `int`
        The channel's creator's identifier. Defaults to `0`.
    parent_id : `int`
        The channel's parent's identifier.
    slowmode : `int`
        The amount of time in seconds what a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages`, `manage_channels` permissions are unaffected.
    thread_users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        The users inside of the thread if any.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('invitable',)
    
    
    def __new__(
        cls,
        *,
        archived = ...,
        archived_at = ...,
        auto_archive_after = ...,
        created_at = ...,
        invitable = ...,
        name = ...,
        parent_id = ...,
        open = ...,
        owner_id = ...,
        slowmode = ...,
    ):
        """
        Creates a new guild public thread channel metadata from the given parameters.
        
        Parameters
        ----------
        archived : `bool`, Optional (Keyword only)
            Whether the thread is archived.
        archived_at : `None | DateTime`, Optional (Keyword only)
            When the thread's archive status was last changed.
        auto_archive_after : `int`, Optional (Keyword only)
            Duration in seconds to automatically archive the thread after recent activity.
        created_at : `None | DateTime`, Optional (Keyword only)
            When the channel was created.
        invitable : `bool`, Optional (Keyword only)
            Whether non-moderators can invite other non-moderators to the threads.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        open : `bool`, Optional (Keyword only)
            Whether the thread channel is open.
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The channel's creator's identifier.
        slowmode : `int`, Optional (Keyword only)
            The amount of time in seconds what a user needs to wait between it's each message.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # invitable
        if invitable is ...:
            invitable = True
        else:
            invitable = validate_invitable(invitable)
        
        # Construct
        self = ChannelMetadataGuildThreadBase.__new__(
            cls,
            archived = archived,
            archived_at = archived_at,
            auto_archive_after = auto_archive_after,
            created_at = created_at,
            name = name,
            open = open,
            owner_id = owner_id,
            parent_id = parent_id,
            slowmode = slowmode,
        )
        self.invitable = invitable
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildThreadBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            archived = keyword_parameters.pop('archived', ...),
            archived_at = keyword_parameters.pop('archived_at', ...),
            auto_archive_after = keyword_parameters.pop('auto_archive_after', ...),
            created_at = keyword_parameters.pop('created_at', ...),
            invitable = keyword_parameters.pop('invitable', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            open = keyword_parameters.pop('open', ...),
            owner_id = keyword_parameters.pop('owner_id', ...),
            slowmode = keyword_parameters.pop('slowmode', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildThreadBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildThreadBase.__hash__(self)
        
        # invitable
        hash_value ^= self.invitable << 9
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildThreadBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildThreadBase._is_equal_same_type(self, other):
            return False
        
        # invitable
        if self.invitable != other.invitable:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildThreadBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildThreadPrivate, cls)._create_empty()
        
        self.invitable = True
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildThreadBase.copy)
    def copy(self):
        new = ChannelMetadataGuildThreadBase.copy(self)
        new.invitable = self.invitable
        return new
    
    
    def copy_with(
        self,
        *,
        archived = ...,
        archived_at = ...,
        auto_archive_after = ...,
        created_at = ...,
        invitable = ...,
        name = ...,
        parent_id = ...,
        open = ...,
        owner_id = ...,
        slowmode = ...,
    ):
        """
        Copies the guild public thread channel metadata with the given fields.
        
        Parameters
        ----------
        archived : `bool`, Optional (Keyword only)
            Whether the thread is archived.
        archived_at : `None | DateTime`, Optional (Keyword only)
            When the thread's archive status was last changed.
        auto_archive_after : `int`, Optional (Keyword only)
            Duration in seconds to automatically archive the thread after recent activity.
        created_at : `None | DateTime`, Optional (Keyword only)
            When the channel was created.
        invitable : `bool`, Optional (Keyword only)
            Whether non-moderators can invite other non-moderators to the threads.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        open : `bool`, Optional (Keyword only)
            Whether the thread channel is open.
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The channel's creator's identifier.
        slowmode : `int`, Optional (Keyword only)
            The amount of time in seconds what a user needs to wait between it's each message.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # invitable
        if invitable is ...:
            invitable = self.invitable
        else:
            invitable = validate_invitable(invitable)
        
        # Construct
        new = ChannelMetadataGuildThreadBase.copy_with(
            self,
            archived = archived,
            archived_at = archived_at,
            auto_archive_after = auto_archive_after,
            created_at = created_at,
            name = name,
            open = open,
            owner_id = owner_id,
            parent_id = parent_id,
            slowmode = slowmode,
        )
        new.invitable = invitable
        return new
    
    
    @copy_docs(ChannelMetadataGuildThreadBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            archived = keyword_parameters.pop('archived', ...),
            archived_at = keyword_parameters.pop('archived_at', ...),
            auto_archive_after = keyword_parameters.pop('auto_archive_after', ...),
            created_at = keyword_parameters.pop('created_at', ...),
            invitable = keyword_parameters.pop('invitable', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            open = keyword_parameters.pop('open', ...),
            owner_id = keyword_parameters.pop('owner_id', ...),
            slowmode = keyword_parameters.pop('slowmode', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildThreadBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildThreadBase._update_attributes(self, data)
        
        # invitable
        self.invitable = parse_invitable(data)
    
    
    @copy_docs(ChannelMetadataGuildThreadBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildThreadBase._difference_update_attributes(self, data)
        
        # invitable
        invitable = parse_invitable(data)
        if (self.invitable != invitable):
            old_attributes['invitable'] = self.invitable
            self.invitable = invitable
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildThreadBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildThreadBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # invitable
        put_invitable(self.invitable, data, defaults, flatten_thread_metadata = not include_internals)
        
        return data



from .fields import *
from .guild_activity_overview_activity import *
from .preinstanced import *


__all__ = (
    *fields.__all__,
    *guild_activity_overview_activity.__all__,
    *preinstanced.__all__,
)



__all__ = (
    'MessageBuilderForumThreadCreate', 'MessageBuilderBase', 'MessageBuilderCreate', 'MessageBuilderEdit',
    'MessageBuilderInteractionComponentEdit', 'MessageBuilderInteractionFollowupCreate',
    'MessageBuilderInteractionFollowupEdit', 'MessageBuilderInteractionResponseCreate',
    'MessageBuilderInteractionResponseEdit', 'MessageBuilderWebhookCreate', 'MessageBuilderWebhookEdit'
)

from scarletio import copy_docs, export

from ...builder.builder_fielded import BuilderFielded

from .conversions import (
    CONVERSION_ALLOWED_MENTIONS, CONVERSION_APPLIED_TAGS, CONVERSION_APPLIED_TAG_IDS, CONVERSION_ATTACHMENTS,
    CONVERSION_AVATAR_URL, CONVERSION_COMPONENTS, CONVERSION_CONTENT, CONVERSION_EMBED, CONVERSION_EMBEDS,
    CONVERSION_ENFORCE_NONCE, CONVERSION_FLAGS, CONVERSION_FORWARD_MESSAGE, CONVERSION_INSTANCE,
    CONVERSION_MESSAGE_REFERENCE_CONFIGURATION, CONVERSION_NAME, CONVERSION_NONCE, CONVERSION_POLL,
    CONVERSION_REPLY_FAIL_FALLBACK, CONVERSION_REPLY_MESSAGE_ID, CONVERSION_SHARED_CLIENT_THEME,
    CONVERSION_SHOW_FOR_INVOKING_USER_ONLY, CONVERSION_SILENT, CONVERSION_STICKER, CONVERSION_STICKERS,
    CONVERSION_STICKER_IDS, CONVERSION_SUPPRESS_EMBEDS, CONVERSION_THREAD_NAME, CONVERSION_TTS,
    CONVERSION_VOICE_ATTACHMENT
)


@export
class MessageBuilderBase(BuilderFielded):
    """
    Base message builder.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    __conversions_default__ = [
        CONVERSION_INSTANCE,
    ]
    
    
    @copy_docs(BuilderFielded._with_positional_parameter_unknown)
    def _with_positional_parameter_unknown(self, value):
        return self._setter_field(CONVERSION_CONTENT, str(value))
    
    
    allowed_mentions = CONVERSION_ALLOWED_MENTIONS
    components = CONVERSION_COMPONENTS
    content = CONVERSION_CONTENT
    embed = CONVERSION_EMBED
    embeds = CONVERSION_EMBEDS
    flags = CONVERSION_FLAGS
    suppress_embeds = CONVERSION_SUPPRESS_EMBEDS


class MessageBuilderCreate(MessageBuilderBase):
    """
    Message builder for ``Client.message_create`` method.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    attachments = CONVERSION_ATTACHMENTS
    enforce_nonce = CONVERSION_ENFORCE_NONCE
    forward_message = CONVERSION_FORWARD_MESSAGE
    message_reference_configuration = CONVERSION_MESSAGE_REFERENCE_CONFIGURATION
    nonce = CONVERSION_NONCE
    poll = CONVERSION_POLL
    reply_fail_fallback = CONVERSION_REPLY_FAIL_FALLBACK
    reply_message_id = CONVERSION_REPLY_MESSAGE_ID
    shared_client_theme = CONVERSION_SHARED_CLIENT_THEME
    silent = CONVERSION_SILENT
    sticker = CONVERSION_STICKER
    sticker_ids = CONVERSION_STICKER_IDS
    stickers = CONVERSION_STICKERS
    tts = CONVERSION_TTS
    voice_attachment = CONVERSION_VOICE_ATTACHMENT


class MessageBuilderEdit(MessageBuilderBase):
    """
    Message builder for ``Client.message_edit`` method.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    attachments = CONVERSION_ATTACHMENTS


class MessageBuilderWebhookCreate(MessageBuilderBase):
    """
    Message builder for ``Client.webhook_message_create``.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    applied_tags = CONVERSION_APPLIED_TAGS
    applied_tag_ids = CONVERSION_APPLIED_TAG_IDS
    attachments = CONVERSION_ATTACHMENTS
    avatar_url = CONVERSION_AVATAR_URL
    name = CONVERSION_NAME
    poll = CONVERSION_POLL
    silent = CONVERSION_SILENT
    thread_name = CONVERSION_THREAD_NAME
    tts = CONVERSION_TTS
    voice_attachment = CONVERSION_VOICE_ATTACHMENT


class MessageBuilderWebhookEdit(MessageBuilderEdit):
    """
    Message builder for ``Client.webhook_message_edit``.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    
class MessageBuilderForumThreadCreate(MessageBuilderBase):
    """
    Message builder for ``Client.forum_thread_create`` method.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    attachments = CONVERSION_ATTACHMENTS
    nonce = CONVERSION_NONCE
    silent = CONVERSION_SILENT
    sticker = CONVERSION_STICKER
    sticker_ids = CONVERSION_STICKER_IDS
    stickers = CONVERSION_STICKERS
    tts = CONVERSION_TTS


class MessageBuilderInteractionComponentEdit(MessageBuilderBase):
    """
    Message builder for ``Client.interaction_component_message_edit``.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()


class MessageBuilderInteractionFollowupCreate(MessageBuilderBase):
    """
    Message builder for ``Client.interaction_followup_message_create` method.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    attachments = CONVERSION_ATTACHMENTS
    poll = CONVERSION_POLL
    show_for_invoking_user_only = CONVERSION_SHOW_FOR_INVOKING_USER_ONLY
    silent = CONVERSION_SILENT
    tts = CONVERSION_TTS
    voice_attachment = CONVERSION_VOICE_ATTACHMENT


class MessageBuilderInteractionFollowupEdit(MessageBuilderEdit):
    """
    Message builder for ``Client.interaction_followup_message_edit``.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()


class MessageBuilderInteractionResponseCreate(MessageBuilderBase):
    """
    Message builder for ``Client.interaction_response_message_create` method.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    poll = CONVERSION_POLL
    show_for_invoking_user_only = CONVERSION_SHOW_FOR_INVOKING_USER_ONLY
    silent = CONVERSION_SILENT
    tts = CONVERSION_TTS


class MessageBuilderInteractionResponseEdit(MessageBuilderEdit):
    """
    Message builder for ``Client.interaction_response_message_edit``.
    
    Attributes
    ----------
    fields : `dict<Conversion, object>`
        The fields to create the message with.
    """
    __slots__ = ()
    
    poll = CONVERSION_POLL
    voice_attachment = CONVERSION_VOICE_ATTACHMENT



__all__ = ('import_plugin',)

from os.path import basename as get_file_name, splitext as split_file_name_and_extension
from threading import current_thread

from scarletio import export, get_last_module_frame

from ....discord import KOKORO

from ..plugin import PLUGINS
from ..plugin_loader import PLUGIN_LOADER


@export
def import_plugin(plugin_name, *variable_names, **keyword_parameters):
    """
    Imports from the given plugin.
    
    Parameters
    ----------
    plugin_name : `str`
        the plugin's name to import.
    
    *variable_names : `str`
        Variables to load from the plugin.
        
        > If no variable is defined, the module is returned instead.

    **keyword_parameters : Keyword parameters
        Additional parameters to create the plugin with.
    
    Returns
    -------
    imported_variables : `object`, `list` of `object`
        The imported variable(s).
    
    Raises
    ------
    ImportError
        - Cannot find the designated plugin.
    RuntimeError
        - If called from non plugin file.
    TypeError
        - If `plugin_name` is not a string.
        - If `variables_to_import` contains a non string.
    ValueError
        - If `plugin_name`'s syntax is incorrect.
        - If `variable_names` contains a non identifier.
    PluginError
        - Any exception raised by the other plugin is funneled.
    """
    if current_thread() is KOKORO:
        raise RuntimeError(
            f'Cannot execute `{import_plugin.__name__}` on `{KOKORO!r}`; plugin_name = {plugin_name!r}. '
            f'You are probably trying to import a plugin for the first time from an async function.'
        )

    # Validate input types
    if not isinstance(plugin_name, str):
        raise TypeError(
            f'plugin_name` can be `str`, got {plugin_name.__class__.__name__}; {plugin_name!r}.'
        )
    
    for variable_name in variable_names:
        if not isinstance(variable_name, str):
            raise TypeError(
                f'`variable_names` can contain only `str`, got {plugin_name.__class__.__name__}; '
                f'{plugin_name!r}; variable_names = {variable_names!r}.'
            )
        
        if not variable_name.isidentifier():
            raise ValueError
        
    
    # Validate `plugin_name` syntax.
    if not plugin_name:
        raise ValueError(
            f'`plugin_name`\'s syntax is incorrect; Cannot be empty string.'
        )
    
    plugin_name_parts = plugin_name.split('.')
    expects_only_identifiers = False
    
    for plugin_name_part in plugin_name_parts:
        if not expects_only_identifiers:
            if not plugin_name_part:
                continue
            
            expects_only_identifiers = True
        
        if not plugin_name_part:
            raise ValueError(
                f'`plugin_name`\'s syntax is incorrect; two dot character cannot follow an identifier; got: '
                f'{plugin_name!r}.'
            )
        
        if not plugin_name_part.isidentifier():
            raise ValueError(
                f'`plugin_name`\'s syntax is incorrect; {plugin_name_part!r} is not an identifier; got: '
                f'{plugin_name!r}.'
            )
    
    if not plugin_name_parts[-1]:
        raise ValueError(
            f'`plugin_name`\'s syntax is incorrect; Cannot end with dot character; got: {plugin_name!r}.'
        )
    
    
    frame = get_last_module_frame()
    if (frame is None):
        spec = None
    else:
        spec = frame.f_globals.get('__spec__', None)
    
    if spec is None:
        raise RuntimeError(
            f'`import_plugin` can be called only from a plugin.'
        )
    
    local_name = spec.name
    
    is_init_file = (split_file_name_and_extension(get_file_name(spec.origin))[0] == '__init__')
    
    empty_count = 0
    for plugin_name_part in plugin_name_parts:
        if plugin_name_part:
            break
        
        empty_count += 1
        continue
    
    if empty_count:
        local_name_parts = local_name.split('.')
        
        if len(local_name_parts) < empty_count:
            raise ImportError(
                f'`{plugin_name!r}`\'s scope out of parent root\'s scope.'
            )
        
        plugin_name_parts = plugin_name_parts[empty_count:]
        
        empty_count = empty_count - is_init_file
        if empty_count:
            local_name_parts = local_name_parts[:-empty_count]
        
        built_name = '.'.join(local_name_parts+plugin_name_parts)
        
    else:
        built_name = plugin_name
    
    plugin = PLUGIN_LOADER.register_and_load(built_name, **keyword_parameters)
    
    current_plugin = PLUGINS.get(local_name, None)
    if (current_plugin is not None):
        if current_plugin.is_directory():
            current_plugin.add_sub_module_plugin(plugin)
        else:
            current_plugin.add_child_plugin(plugin)
            plugin.add_parent_plugin(current_plugin)
        
    variable_names_length = len(variable_names)
    module = plugin._spec.get_module_proxy()
    if variable_names_length == 0:
        return module
    
    variables = []
    
    for variable_name in variable_names:
        try:
            variable = getattr(module, variable_name)
        except AttributeError:
            raise ImportError(
                f'Cannot import `{variable_name}` from `{plugin.name}`.'
            ) from None
        
        variables.append(variable)
    
    if variable_names_length == 1:
        return variables[0]
    
    return variables



__all__ = ('ScheduledEventEntityMetadataBase',)

from scarletio import RichAttributeErrorBaseType

from ...bases import PlaceHolder
from ...user import create_partial_user_from_id


class ScheduledEventEntityMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for ``ScheduledEvent``'s entity metadata.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new entity metadata instance.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_keyword_parameters(cls, keyword_parameters):
        """
        Creates a new entity metadata instance from the given keyword_parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Additional keyword parameters passed to ``ScheduledEvent.__new__``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return cls()
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new scheduled event entity metadata instance from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity metadata structure.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the entity metadata to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        return {}
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates a scheduled event entity metadat instance with its attributes set with their default values.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the entity metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def __hash__(self):
        """Returns the entity metadata's hash value."""
        return 0
    
    
    def __eq__(self, other):
        """Returns whether the two entity metadatas equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two entity metadatas are equal.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        return True
    
    
    def copy(self):
        """
        Copies the scheduled event entity metadata.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the scheduled event entity metadata with the given fields.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return object.__new__(type(self))
    
    
    def copy_with_keyword_parameters(self, keyword_parameters):
        """
        Copies the scheduled event entity metadata with the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Additional keyword parameters passed to ``ScheduledEvent.copy_with``
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return self.copy_with()
    
    
    location = PlaceHolder(
        None,
        """
        Returns the place where the event will take place.
        
        Returns
        -------
        location : `None`, `str`
        """
    )
    
    
    speaker_ids = PlaceHolder(
        None,
        """
        Returns the speakers' identifier of the stage channel.
        
        Returns
        -------
        speaker_ids : `None | tuple<int>`
        """
    )
    
    
    def iter_speaker_ids(self):
        """
        Iterates over the speakers' identifiers of the stage channel.
        
        This method is an iterable generator.
        
        Yields
        ------
        speaker_id : `int`
        """
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            yield from speaker_ids
    
    
    @property
    def speakers(self):
        """
        Returns the speakers of the stage channel.
        
        Returns
        -------
        speakers : `None`, `tuple` of ``ClientUserBase``
        """
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            return (*(create_partial_user_from_id(speaker_id) for speaker_id in speaker_ids),)
    
    
    def iter_speakers(self):
        """
        Iterates over the speakers of the stage channel.
        
        This method is an iterable generator.
        
        Yields
        ------
        speakers : ``ClientUserBase``
        """
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            
            for speaker_id in speaker_ids:
                yield create_partial_user_from_id(speaker_id)



__all__ = ()

from ...soundboard.soundboard_sound.fields import (
    validate_available, validate_emoji, validate_id, validate_name, validate_user_id, validate_volume
)

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..audit_log_entry_change_conversion.change_deserializers import (
    change_deserializer_deprecation, change_deserializer_flattened_emoji
)
from ..audit_log_entry_change_conversion.change_serializers import change_serializer_flattened_emoji
from ..audit_log_entry_change_conversion.value_mergers import value_merger_replace
from ..conversion_helpers.converters import (
    value_deserializer_id, value_deserializer_name, value_serializer_id, value_serializer_name
)


# ---- available ----

AVAILABLE_CONVERSION = AuditLogEntryChangeConversion(
    ('available',),
    'available',
    value_validator = validate_available,
)


@AVAILABLE_CONVERSION.set_value_deserializer
def available_value_deserializer(value):
    if value is None:
        value = True
    return value


# ---- emoji ----

EMOJI_CONVERSION = AuditLogEntryChangeConversion(
    ('emoji_id', 'emoji_name'),
    'emoji',
    change_deserializer = change_deserializer_flattened_emoji,
    change_serializer = change_serializer_flattened_emoji,
    value_validator = validate_emoji,
    value_merger = value_merger_replace
)

# ---- id  ----

ID_CONVERSION = AuditLogEntryChangeConversion(
    ('sound_id',),
    'id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_id,
)


ID_CONVERSION_IGNORED = AuditLogEntryChangeConversion(
    ('id',),
    '',
    change_deserializer = change_deserializer_deprecation,
)

# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)

# ---- user_id ----

USER_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('user_id',),
    'user_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_user_id,
)


# ---- volume ----

VOLUME_CONVERSION = AuditLogEntryChangeConversion(
    ('volume',),
    'volume',
    value_validator = validate_volume,
)


@VOLUME_CONVERSION.set_value_deserializer
def volume_value_deserializer(value):
    if (value is None):
        value = 1.0
    
    return value


# ---- Construct ----

SOUNDBOARD_SOUND_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    AVAILABLE_CONVERSION,
    EMOJI_CONVERSION,
    ID_CONVERSION,
    ID_CONVERSION_IGNORED,
    NAME_CONVERSION,
    USER_ID_CONVERSION,
    VOLUME_CONVERSION,
)



__all__ = ()

from ...utils import seconds_to_id_difference

from ..interaction_metadata import InteractionMetadataBase


INTERACTION_EVENT_EXPIRE_AFTER = 900 # 15 min
INTERACTION_EVENT_EXPIRE_AFTER_ID_DIFFERENCE = seconds_to_id_difference(INTERACTION_EVENT_EXPIRE_AFTER)

DEFAULT_INTERACTION_METADATA = InteractionMetadataBase()



__all__ = ('create_project_structure',)

from os import mkdir as create_directory
from os.path import join as join_paths

from ....... import __package__ as PACKAGE_NAME

from ...helpers import create_directory_recursive
from ...naming import get_bot_constant_name, get_bot_module_name, get_bot_variable_name

from .readme_rendering import build_readme_content


def create_file(directory_path, file_name, content):
    """
    Creates a file at the given position described by `directory_path` and `file_name`.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    file_name : `str`
        The file's name.
    content : `str`
        The content to write.
    """
    with open(join_paths(directory_path, file_name), 'w', encoding = 'utf-8') as file:
        file.write(content)


def create_gitignore_file(directory_path):
    """
    Creates a `/{project_name}/.gitignore` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    """
    create_file(
        directory_path,
        '.gitignore',
        (
            f'# Generic\n'
            f'*.env\n'
            f'*.venv\n'
            f'\n'
            f'# Python\n'
            f'*.pyc\n'
            f'__pycache__/\n'
            f'\n'
            f'# Editor\n'
            f'.idea\n'
            f'\n'
            f'# Data files\n'
            f'*.pickle\n'
            f'*.sqlite3\n'
            f'\n'
            f'# Hata\n'
            f'.profiles/\n'
        ),
    )


def create_readme_file(directory_path, project_name, bot_names):
    """
    Creates a `/README.md` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    project_name : `str`
        The project's name.
    bot_names : `list` of `str`
        The bots' names.
    """
    create_file(directory_path, 'README.md', build_readme_content(project_name, bot_names))


def create_pyproject_toml_file(directory_path, project_name):
    """
    Creates a `/pyproject.toml` file`.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    
    project_name : `str`
        The project's name.
    """
    create_file(
        directory_path,
        'pyproject.toml',
        (
            f'[project]\n'
            f'name = \"{project_name}\"\n'
            f'\n'
            f'dependencies = [\n'
            f'    \"{PACKAGE_NAME!s}\",\n'
            f'    \"{PACKAGE_NAME!s}[all]\",\n'
            f']\n'
            f'readme.file = "README.md"\n'
            f'readme.content-type = "text/markdown"\n'
            f'requires-python = \">=3.6\"\n'
            f'\n'
            f'dynamic = [\n'
            f'    \"version\",\n'
            f'    \"optional-dependencies\",\n'
            f']\n'
            f'\n'
            f'[build-system]\n'
            f'build-backend = \"setuptools.build_meta\"\n'
            f'requires = [\n'
            f'    \"setuptools\",\n'
            f'    \"setuptools-scm\"\n'
            f']\n'
            f'\n'
            f'[project.scripts]\n'
            f'# Allows doing: "$ {project_name!s}" from terminal after installed.\n'
            f'{project_name!s} = \"{project_name!s}.cli:main\"\n'
            f'\n'
            f'[tool.setuptools]\n'
            f'include-package-data = false\n'
            f'\n'
            f'# `packages` are the directories with that should be included when installed.\n'
            f'# Includes all the `.py` files by default and no other files.\n'
            f'# Should NOT include test directories.\n'
            f'packages = [\n'
            f'    \"{project_name!s}\",\n'
            f'    \"{project_name!s}.bots\",\n'
            f'    \"{project_name!s}.plugins\",\n'
            f']\n'
            f'\n'
            f'[tool.setuptools.package-data]\n'
            f'# Additional files that should be included when installing\n'
            f'# Example: include all `.png` files within `images` plugin\'s assets directory\n'
            f'# This directory also have to be added to `packages` to work.\n'
            f'# (this is not an actually created plugin, just an example):\n'
            f'# "{project_name!s}.plugins.images.assets" = ["*.png"]\n'
            f'# Example: include all `.txt` files:\n'
            f'# "*" = ["*.txt"]\n'
            f'\n'
            f'# References:\n'
            f'# - https://setuptools.pypa.io/en/latest/userguide/pyproject_config.html\n'
            f'# - https://setuptools.pypa.io/en/latest/userguide/datafiles.html\n'
        ),
    )


def build_dot_env_file_content(bot_names):
    """
    Builds the content of `/{project_name}/.env` file.
    
    Parameters
    ----------
    bot_names : `list` of `str`
        The bots' names.
    
    Returns
    -------
    content : `str`
    """
    content_parts = []
    
    for bot_name in bot_names:
        bot_constant_name = get_bot_constant_name(bot_name)
        content_parts.append(bot_constant_name)
        content_parts.append('_TOKEN =\n')
    
    return ''.join(content_parts)


def create_dot_env_file(directory_path, bot_names):
    """
    Creates a `/{project_name}/.env` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    bot_names : `list` of `str`
        The bots' names.
    """
    create_file(directory_path, '.env', build_dot_env_file_content(bot_names))


def create_project_init_file(directory_path):
    """
    Creates a `/{project_name}/__init__.py` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    """
    create_file(directory_path, '__init__.py', '__all__ = ()\n')


def create_cli_file(directory_path):
    """
    Creates a `/{project_name}/cli.py` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    """
    create_file(
        directory_path,
        'cli.py',
        (
            f'__all__ = (\'main\',)\n'
            f'\n'
            f'def main():\n'
            f'    try:\n'
            f'        from {PACKAGE_NAME!s}.main import execute_command_from_system_parameters\n'
            f'    except ImportError as err:\n'
            f'        raise ImportError(\n'
            f'            \'Couldn\\\'t import {PACKAGE_NAME!s}. \'\n'
            f'            \'Are you sure it\\\'s installed and available on your PYTHONPATH environment variable? \'\n'
            f'            \'Did you forget to activate a virtual environment?\'\n'
            f'        ) from err\n'
            f'\n'
            f'    from {PACKAGE_NAME!s}.ext.plugin_auto_reloader import start_auto_reloader, warn_auto_reloader_availability\n'
            f'    from {PACKAGE_NAME!s}.ext.plugin_loader import load_all_plugin, frame_filter, register_plugin\n'
            f'    from scarletio import get_event_loop, write_exception_sync\n'
            f'\n'
            f'    from . import bots\n'
            f'\n'
            f'    register_plugin(f\'{{__spec__.parent}}.plugins\')\n'
            f'\n'
            f'    try:\n'
            f'        load_all_plugin()\n'
            f'    except BaseException as err:\n'
            f'        write_exception_sync(err, filter = frame_filter)\n'
            f'        get_event_loop().stop()\n'
            f'        raise SystemExit(1) from None\n'
            f'\n'
            f'    warn_auto_reloader_availability()\n'
            f'    start_auto_reloader()\n'
            f'\n'
            f'    execute_command_from_system_parameters()\n'
        ),
    )


def build_constants_file_content(bot_names):
    """
    Builds the content of the `/{project_name}/constants.py` file.
    
    Parameters
    ----------
    bot_names : `list` of `str`
        The bots' names.
    
    Returns
    -------
    content : `str`
    """
    content_parts = []
    
    # Header
    content_parts.append('__all__ = ()\n\n')
    
    # Imports
    content_parts.append('from ')
    content_parts.append(PACKAGE_NAME)
    content_parts.append('.env import EnvGetter\n\n\n')
    
    # Content
    content_parts.append('with EnvGetter() as env:\n')
    for bot_name in bot_names:
        constant_name = get_bot_constant_name(bot_name)
        
        content_parts.append('    ')
        content_parts.append(constant_name)
        content_parts.append('_TOKEN = env.get_str(\'')
        content_parts.append(constant_name)
        content_parts.append('_TOKEN\', raise_if_missing_or_empty = True)\n')
    
    return ''.join(content_parts)


def create_constants_file(directory_path, bot_names):
    """
    Creates a new `/{project_name}/constants.py` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    bot_names : `list` of `str`
        The bots' names.
    """
    create_file(directory_path, 'constants.py', build_constants_file_content(bot_names))


def create_main_file(directory_path):
    """
    Creates a `/{project_name}/__main__.py` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    """
    create_file(
        directory_path,
        '__main__.py',
        (
            f'__all__ = ()\n'
            f'\n'
            f'from .cli import main\n'
            f'\n'
            f'if __name__ == \'__main__\':\n'
            f'    main()\n'
        ),
    )


def build_bots_init_file_content(bot_names):
    """
    Builds the content of `/{project_name}/bots/__init__.py` file.
    
    Parameters
    ----------
    bot_names : `list` of `str`
        The bots' names.
    
    Returns
    -------
    content : `str`
    """
    content_parts = []
    
    bot_module_names = [get_bot_module_name(bot_name) for bot_name in bot_names]
    
    for bot_module_name in bot_module_names:
        content_parts.append('from .')
        content_parts.append(bot_module_name)
        content_parts.append(' import *\n')
    
    content_parts.append('\n\n__all__ = (\n')
    
    for bot_module_name in bot_module_names:
        content_parts.append('    *')
        content_parts.append(bot_module_name)
        content_parts.append('.__all__,\n')
    
    content_parts.append(')\n')
    
    return ''.join(content_parts)


def create_bots_init_file(directory_path, bot_names):
    """
    Creates a `/{project_name}/bots/__init__.py` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    bot_names : `list` of `str`
        The bots' names.
    """
    create_file(directory_path, '__init__.py', build_bots_init_file_content(bot_names))


def create_bot_file(directory_path, bot_name):
    """
    Creates a `/{project_name}/bots/{bot_name}.py` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    bot_name : `str`
        The bot's name.
    """
    bot_module_name = get_bot_module_name(bot_name)
    bot_variable_name = get_bot_variable_name(bot_name)
    bot_constant_name = get_bot_constant_name(bot_name)
    
    create_file(
        directory_path,
        bot_module_name + '.py',
        (
            f'__all__ = (\'{bot_variable_name}\',)\n'
            f'\n'
            f'from {PACKAGE_NAME!s} import Client\n'
            f'\n'
            f'from ..constants import {bot_constant_name}_TOKEN\n'
            f'\n'
            f'\n'
            f'{bot_variable_name} = Client(\n'
            f'    {bot_constant_name}_TOKEN,\n'
            f'    extensions = [\'slash\'],\n'
            f')\n'
        ),
    )


def create_plugins_init_file(directory_path):
    """
    Creates a `/{project_name}/plugins/__init__.py` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    """
    create_file(
        directory_path,
        '__init__.py',
        (
            f'from {PACKAGE_NAME!s}.ext.plugin_loader import mark_as_plugin_root_directory\n'
            f'\n'
            f'mark_as_plugin_root_directory()\n'
        ),
    )


def create_plugins_ping_command_file(directory_path):
    """
    Creates a `/{project_name}/plugins/ping.py` file.
    
    Parameters
    ----------
    directory_path : `str`
        Path to the file's directory.
    """
    create_file(
        directory_path,
        'ping.py',
        (
            f'from time import perf_counter\n'
            f'\n'
            f'from {PACKAGE_NAME!s} import ClientWrapper\n'
            f'\n'
            f'\n'
            f'ALL = ClientWrapper()\n'
            f'\n'
            f'\n'
            f'@ALL.interactions(is_global = True, wait_for_acknowledgement = True)\n'
            f'async def ping():\n'
            f'    \"\"\"HTTP ping-pong.\"\"\"\n'
            f'    start = perf_counter()\n'
            f'    yield\n'
            f'    delay = (perf_counter() - start) * 1000.0\n'
            f'\n'
            f'    yield f\'{{delay:.0f}} ms\'\n'
        ),
    )


def create_project_structure(root_directory_path, project_name, bot_names):
    """
    Creates a project with package layout.
    
    Parameters
    ----------
    root_directory_path : `str`
        The project's root directory's path.
    project_name : `str`
        The project's name.
    bot_names : `list` of `str`
        The bots' names.
    """
    # root
    create_directory_recursive(root_directory_path)
    
    create_gitignore_file(root_directory_path)
    create_readme_file(root_directory_path, project_name, bot_names)
    create_pyproject_toml_file(root_directory_path, project_name)
    
    # root / { project_name }
    project_directory_path = join_paths(root_directory_path, project_name)
    create_directory(project_directory_path)
    
    create_project_init_file(project_directory_path)
    create_main_file(project_directory_path)
    create_dot_env_file(project_directory_path, bot_names)
    create_cli_file(project_directory_path)
    create_constants_file(project_directory_path, bot_names)
    
    # root / { project_name } / bots
    bots_directory_path = join_paths(project_directory_path, 'bots')
    create_directory(bots_directory_path)
    
    create_bots_init_file(bots_directory_path, bot_names)
    for bot_name in bot_names:
        create_bot_file(bots_directory_path, bot_name)
    
    # root / { project_name } / plugins
    plugins_directory_path = join_paths(project_directory_path, 'plugins')
    create_directory(plugins_directory_path)
    
    create_plugins_init_file(plugins_directory_path)
    create_plugins_ping_command_file(plugins_directory_path)



__all__ = ()

from ...field_parsers import nullable_string_parser_factory
from ...field_putters import nullable_string_putter_factory, url_optional_putter_factory
from ...field_validators import url_optional_validator_factory

from .constants import NAME_LENGTH_MAX, URL_LENGTH_MAX

# name

parse_name = nullable_string_parser_factory('name')
put_name = nullable_string_putter_factory('name')


def validate_name(name):
    """
    Validates the given embed provider name.
    
    Parameters
    ----------
    name : `None`, `str`, `object`
        Embed provider name.
    
    Returns
    -------
    name : `None`, `str`
    
    Raises
    ------
    TypeError
        - If `name`'s type is incorrect.
    ValueError
        - If `name`'s length is out of the expected range.
    """
    if name is None:
        return None
    
    if not isinstance(name, str):
        name = str(name)
    
    name_length = len(name)
    if name_length == 0:
        return None
    
    if name_length > NAME_LENGTH_MAX:
        raise ValueError(
            f'`name` length` must be <= {NAME_LENGTH_MAX}, got {name_length}; name = {name!r}.'
        )
    
    return name

# url

parse_url = nullable_string_parser_factory('url')
put_url = url_optional_putter_factory('url')
validate_url = url_optional_validator_factory('url', length_max = URL_LENGTH_MAX)



from .allowed_mentions import *
from .applied_tag_ids import *
from .applied_tags import *
from .attachments import *
from .avatar_url import *
from .components import *
from .content import *
from .embed import *
from .embeds import *
from .enforce_nonce import *
from .flags import *
from .forward_message import *
from .instance import *
from .message_reference_configuration import *
from .name import *
from .nonce import *
from .poll import *
from .reply_fail_fallback import *
from .reply_message_id import *
from .shared_client_theme import *
from .show_for_invoking_user_only import *
from .silent import *
from .sticker import *
from .sticker_ids import *
from .stickers import *
from .suppress_embeds import *
from .thread_name import *
from .tts import *
from .voice_attachment import *


__all__ = ()



__all__ = ('NamePlate',)

from ...bases import Slotted
from ...http.urls import build_name_plate_url, build_name_plate_url_as
from ...utils import DATETIME_FORMAT_CODE

from .fields import (
    parse_asset_path, parse_name, parse_expires_at, parse_palette, parse_sku_id, put_asset_path, put_expires_at,
    put_name, put_palette, put_sku_id, validate_asset_path, validate_expires_at, validate_name, validate_palette,
    validate_sku_id
)
from .preinstanced import Palette


class NamePlate(metaclass = Slotted):
    """
    Represents a user's name's plate in a user listing.
    
    Attributes
    ----------
    asset_path : `str`
        Middle part of the path to the name plate's asset.
    
    expires_at : `None | DateTime`
        When the name plate expires.
    
    name : `str`
        The name plate's system name.
    
    palette : ``Palette``
        The dominant color of the name plate.
    
    sku_id : `int`
        The stock keeping unit the name plate is part of.
    """
    __slots__ = ('asset_path', 'expires_at', 'name', 'palette', 'sku_id',)
    
    def __new__(cls, *, asset_path = ..., expires_at = ..., name = ..., palette = ..., sku_id = ...):
        """
        Creates a new user name plate instance from the given parameters.
        
        Attributes
        ----------
        asset_path : `None | str`, Optional (Keyword only)
            Middle part of the path to the name plate's asset.
        
        expires_at : `None | DateTime`, Optional (Keyword only)
            When the name plate expires.
        
        name : `None | str`, Optional (Keyword only)
            The name plate's system name.
        
        palette : ``None | str | Palette``, Optional (Keyword only)
            The dominant color of the name plate.
        
        sku_id : `None | int`, Optional (Keyword only)
            The stock keeping unit the name plate is part of.
        """
        # asset_path
        if asset_path is ...:
            asset_path = ''
        else:
            asset_path = validate_asset_path(asset_path)
        
        # expires_at
        if expires_at is ...:
            expires_at = None
        else:
            expires_at = validate_expires_at(expires_at)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # palette
        if palette is ...:
            palette = Palette.black
        else:
            palette = validate_palette(palette)
        
        # sku_id
        if sku_id is ...:
            sku_id = 0
        else:
            sku_id = validate_sku_id(sku_id)
        
        # Construct
        self = object.__new__(cls)
        self.asset_path = asset_path
        self.expires_at = expires_at
        self.name = name
        self.palette = palette
        self.sku_id = sku_id
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # asset_path
        repr_parts.append(' asset_path = ')
        repr_parts.append(repr(self.asset_path))
        
        # expires_at
        expires_at = self.expires_at
        if (expires_at is not None):
            repr_parts.append(', expires_at = ')
            repr_parts.append(format(expires_at, DATETIME_FORMAT_CODE))
        
        # name
        repr_parts.append(', name = ')
        repr_parts.append(repr(self.name))
        
        # palette
        palette = self.palette
        repr_parts.append(', palette = ')
        repr_parts.append(palette.name)
        repr_parts.append(' ~ ')
        repr_parts.append(palette.value)
        
        # sku_id
        repr_parts.append(', sku_id = ')
        repr_parts.append(repr(self.sku_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # asset_path
        hash_value ^= hash(self.asset_path)
        
        # expires_at
        expires_at = self.expires_at
        if (expires_at is not None):
            hash_value ^= hash(expires_at)
        
        # name
        hash_value ^= hash(self.name)
        
        # palette
        hash_value ^= hash(self.palette)
        
        # sku_id
        hash_value ^= self.sku_id
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # asset_path
        if self.asset_path != other.asset_path:
            return False
        
        # expires_at
        if self.expires_at != other.expires_at:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # palette
        if self.palette != other.palette:
            return False
        
        # sku_id
        if self.sku_id != other.sku_id:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a name plate from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received name plate data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.asset_path = parse_asset_path(data)
        self.expires_at = parse_expires_at(data)
        self.name = parse_name(data)
        self.palette = parse_palette(data)
        self.sku_id = parse_sku_id(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the name plate to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_asset_path(self.asset_path, data, defaults)
        put_expires_at(self.expires_at, data, defaults)
        put_name(self.name, data, defaults)
        put_palette(self.palette, data, defaults)
        put_sku_id(self.sku_id, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Copies the name plate.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.asset_path = self.asset_path
        new.expires_at = self.expires_at
        new.name = self.name
        new.palette = self.palette
        new.sku_id = self.sku_id
        return new
    
    
    def copy_with(
        self, 
        *,
        asset_path = ...,
        expires_at = ...,
        name = ...,
        palette = ...,
        sku_id = ...,
    ):
        """
        Copies the name plate with the given fields.
        
        Parameters
        ----------
        asset_path : `None | str`, Optional (Keyword only)
            Middle part of the path to the name plate's asset.
        
        expires_at : `None | DateTime`, Optional (Keyword only)
            When the name plate expires.
        
        name : `None | str`, Optional (Keyword only)
            The name plate's system name.
        
        palette : ``None | str | Palette``, Optional (Keyword only)
            The dominant color of the name plate.
        
        sku_id : `None | int`, Optional (Keyword only)
            The stock keeping unit the name plate is part of.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # asset_path
        if asset_path is ...:
            asset_path = self.asset_path
        else:
            asset_path = validate_asset_path(asset_path)
        
        # expires_at
        if expires_at is ...:
            expires_at = self.expires_at
        else:
            expires_at = validate_expires_at(expires_at)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # palette
        if palette is ...:
            palette = self.palette
        else:
            palette = validate_palette(palette)
        
        # sku_id
        if sku_id is ...:
            sku_id = self.sku_id
        else:
            sku_id = validate_sku_id(sku_id)
        
        # Construct
        new = object.__new__(type(self))
        new.asset_path = asset_path
        new.expires_at = expires_at
        new.name = name
        new.palette = palette
        new.sku_id = sku_id
        return new
    
    
    @property
    def url(self):
        """
        Returns the name plate's asset's url.
        
        Returns
        -------
        url : `str`
        """
        return build_name_plate_url(self.asset_path)
    
    
    def url_as(self, animated = True):
        """
        Returns the name plate's asset's url.
        
        Parameters
        ----------
        animated : `bool` = `False`, Optional
            Whether to return an url to an animated asset.
        
        Returns
        -------
        url : `str`
        """
        return build_name_plate_url_as(self.asset_path, animated)



__all__ = ()

from ...field_parsers import default_entity_parser_factory, preinstanced_parser_factory
from ...field_putters import default_entity_putter_factory, preinstanced_putter_factory
from ...field_validators import default_entity_validator_factory, preinstanced_validator_factory
from ...user import ClientUserBase, User, ZEROUSER

from .preinstanced import TeamMemberRole, TeamMembershipState

# role

parse_role = preinstanced_parser_factory('role', TeamMemberRole, TeamMemberRole.none)
put_role = preinstanced_putter_factory('role')
validate_role = preinstanced_validator_factory('role', TeamMemberRole)

# state

parse_state = preinstanced_parser_factory('membership_state', TeamMembershipState, TeamMembershipState.none)
put_state = preinstanced_putter_factory('membership_state')
validate_state = preinstanced_validator_factory('state', TeamMembershipState)

# user

parse_user = default_entity_parser_factory('user', User, default = ZEROUSER)
put_user = default_entity_putter_factory('user', ClientUserBase, ZEROUSER, force_include_internals = True)
validate_user = default_entity_validator_factory('user', ClientUserBase, default = ZEROUSER)
